import logging
import signal
import time
import docker
import uuid

from .ping import ping

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ServerInstanceHandler:
    def signal_handler(self, sig, frame):
        self.stop()

    def wait(self, timeout: int = 120):
        """Timeout is in seconds"""

        logging.info(f"Waiting for server to be up ...")

        for n in range(timeout):
            logging.debug(f"Trying to ping localhost:{self.port}")

            try:
                latency = ping("localhost", self.port)
                logging.info(f"Success ping in {latency} after {n}s")
                return latency
            except:
                logging.debug(f"Failed status ping {n}")

            time.sleep(1)

        logging.info(f"Couldn't ping the server")
        self.stop()

    def stop(self):
        raise NotImplementedError


class DockerInstance(ServerInstanceHandler):
    def __init__(
        self,
        port: int = 25565,
        rcon_port: int = 25575,
        container_name: str = uuid.uuid4(),
        wait: bool = True,
        image: str = "ghcr.io/portalhubyt/template_server:latest",
        eula: bool = False,
        version: str = "latest",
    ):

        if eula == False:
            exit(
                "Please accept the EULA before running the server by providing 'eula=True' as an argument to create()"
            )

        signal.signal(signal.SIGINT, self.signal_handler)
        self.name = container_name
        self.port = port
        self.rcon_port = rcon_port
        self.eula = eula
        self.version = version
        self.client = docker.APIClient()
        self.client.pull(image)

        try:
            self.container = self.client.create_container(
                image,
                ports=[25565, 25575],
                host_config=self.client.create_host_config(
                    port_bindings={25565: port, 25575: self.rcon_port}
                ),
                environment=[f"EULA={self.eula}", f"VERSION={self.version}"],
                name=container_name,
            )

            self.client.start(self.container)
        except Exception as e:
            raise e
        if wait == True:
            self.wait()

    def stop(self):

        logging.info(f"Stopping ...")
        self.client.stop(self.container)
        self.client.wait(self.container)
        self.client.remove_container(self.container)
        logging.info(f"Removed container")



class BlockState:
    def __init__(self, properties: dict = None):

        if isinstance(properties, dict):
            for key in properties:
                setattr(self, key, properties[key])

        elif properties is not None:
            raise ValueError("BlockState arg must be a dictionnary")

    def flatten(self, arg):

        if isinstance(arg, bool):
            if arg == True:
                return "true"
            else:
                return "false"

        elif isinstance(arg, int):
            return str(arg)

        elif isinstance(arg, str):
            return arg

        else:
            raise UnexpectedBlockStatePropertyValueType(
                f"The type of: [{arg}] is [{type(arg)}] and is not valid for blockstate. The value must either be an int, a boolean or a string"
            )

    def __str__(self):

        buff = ""

        for key in dir(self):
            if key.startswith("__"):
                continue
            elif key == "flatten":
                continue

            value = getattr(self, key)
            value = self.flatten(value)
            buff += f"{key}={value},"

        if buff != "":
            buff = buff[:-1]
            return f"[{buff}]"
        else:
            return ""


class UnexpectedBlockStatePropertyValueType(Exception):
    pass



from typing import Union

from .NBT import NBT
from .TargetSelector import TargetSelector


class Entity:
    """
    Must be a player name, a target selector or a UUID.

    Each entity argument may place limits on the number of
    entities (single/multiple) selected or the type of entities
    (player/any entity) selected.

    e.g:
        - Player
        - 0123
        - @e
        - @e[type=foo]
        - dd12be42-52a9-4a91-a8a1-11c01849e498

    An instance of this class represents an existing unique entity, or a group of unique entities.
    """

    def __init__(self, descriptor: Union[TargetSelector, str]):

        if not isinstance(descriptor, TargetSelector) and not isinstance(
            descriptor, str
        ):
            raise IncorrectEntityDescriptorType(
                f"The Entity descriptor must either be a TargetSelector instance or a string"
            )

        if not descriptor:
            raise ValueError(
                f"The descriptor provided to to TargetSelector was empty or invalid"
            )

        self.descriptor = str(descriptor)

    def __str__(self):
        return f"{self.descriptor}"


class IncorrectEntityDescriptorType(Exception):
    pass



from typing import Union

from .BlockCoordinates import BlockCoordinates


class Zone:
    """
    Custom component that is a set of two BlockCoordinates(), representing
    a rectangular area
    """

    def __init__(
        self, pos1: Union[BlockCoordinates, tuple], pos2: Union[BlockCoordinates, tuple]
    ):

        if isinstance(pos1, tuple):
            self.pos1 = BlockCoordinates(*pos1)
        else:
            self.pos1 = pos1

        if isinstance(pos2, tuple):
            self.pos2 = BlockCoordinates(*pos2)
        else:
            self.pos2 = pos2

    def __str__(self):
        if not isinstance(self.pos1, BlockCoordinates) or not isinstance(
            self.pos2, BlockCoordinates
        ):
            raise ZoneWrongCoordsType(
                "a Zone requires to be provided a set of two BlockCoordinates instances as pos1 and pos2"
            )

        return f"{self.pos1} {self.pos2}"


class ZoneWrongCoordsType(Exception):
    pass



from typing import Union


class Coordinates:
    def __init__(
        self,
        x: Union[int, float, str],
        y: Union[int, float, str],
        z: Union[int, float, str],
    ):

        self.x = x
        self.y = y
        self.z = z

    def check_carets(self):
        coords = [self.x, self.y, self.z]

        if any(str(coord).startswith("^") for coord in coords):
            if not all(str(coord).startswith("^") for coord in coords):
                raise WrongCaretNotation(
                    "When using caret notation, every coordinates must start with a caret"
                )

    def __str__(self):
        self.check_carets()

        return f"{self.x} {self.y} {self.z}"


class WrongCaretNotation(Exception):
    pass



import unittest

from shulker.components.TargetSelector import TargetSelector
from shulker.components.TargetSelector import (
    IncorrectTargetSelectorIdentifier,
    IncorrectTargetSelectorArgumentsType,
    InvalidTargetSelectorArgumentKey,
)


class TestBlock(unittest.TestCase):
    def test_targetselector_p_option(self):
        ts = TargetSelector("p")

        diff = str(ts)
        test = "@p[]"
        self.assertEqual(diff, test)

    def test_targetselector_a_option(self):
        ts = TargetSelector("a")

        diff = str(ts)
        test = "@a[]"
        self.assertEqual(diff, test)

    def test_targetselector_r_option(self):
        ts = TargetSelector("r")

        diff = str(ts)
        test = "@r[]"
        self.assertEqual(diff, test)

    def test_targetselector_s_option(self):
        ts = TargetSelector("s")

        diff = str(ts)
        test = "@s[]"
        self.assertEqual(diff, test)

    def test_targetselector_e_option(self):
        ts = TargetSelector("e")

        diff = str(ts)
        test = "@e[]"
        self.assertEqual(diff, test)

    def test_targetselector_invalid_option(self):
        with self.assertRaises(IncorrectTargetSelectorIdentifier):
            TargetSelector("f")

    def test_targetselector_with_no_arguments(self):
        ts = TargetSelector("p", {})

        diff = str(ts)
        test = "@p[]"
        self.assertEqual(diff, test)

    def test_targetselector_with_argument(self):
        ts = TargetSelector("p", {"team": "french"})

        diff = str(ts)
        test = "@p[team=french]"
        self.assertEqual(diff, test)

    def test_targetselector_with_arguments(self):
        ts = TargetSelector("p", {"team": "french", "tag": "baguette"})

        diff = str(ts)
        test = "@p[team=french,tag=baguette]"
        self.assertEqual(diff, test)

    def test_targetselector_with_arguments_listed(self):
        ts = TargetSelector("p", {"team": "french", "tag": ["baguette", "farine"]})

        diff = str(ts)
        test = "@p[team=french,tag=baguette,tag=farine]"
        self.assertEqual(diff, test)

    def test_targetselector_with_invalid_arguments(self):
        with self.assertRaises(InvalidTargetSelectorArgumentKey):
            str(TargetSelector("p", {"totem": 5, "tag": "baguette"}))

    def test_targetselector_with_invalid_argument_type(self):
        with self.assertRaises(IncorrectTargetSelectorArgumentsType):
            str(TargetSelector("p", ["totem", "tag", 5]))


if __name__ == "__main__":
    unittest.main()



import unittest

from shulker.components.BlockState import (
    BlockState,
    UnexpectedBlockStatePropertyValueType,
)
from shulker.components.Block import Block


class TestBlock(unittest.TestCase):
    def test_blockstate_without_args(self):
        bs = BlockState()

        diff = str(bs)
        test = ""
        self.assertEqual(diff, test)

    def test_blockstate_without_attrs(self):
        b = Block("dirt")

        diff = str(b.blockstate)
        test = ""
        self.assertEqual(diff, test)

    def test_blockstate_without_args_passed_to_Block(self):
        bs = BlockState()
        b = Block("dirt", blockstate=bs)

        diff = str(b.blockstate)
        test = ""
        self.assertEqual(diff, test)

    def test_blockstate_without_args_attributed_to_Block(self):
        bs = BlockState()
        b = Block("dirt")

        b.blockstate = bs

        diff = str(b.blockstate)
        test = ""
        self.assertEqual(diff, test)

    def test_blockstate_with_arg(self):
        bs = BlockState({"snowy": True})

        diff = str(bs)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_with_args(self):
        bs = BlockState({"snowy": True, "facing": "north"})

        diff = str(bs)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attr(self):
        bs = BlockState()
        bs.snowy = True

        diff = str(bs)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attrs(self):
        bs = BlockState()
        bs.snowy = True
        bs.facing = "north"

        diff = str(bs)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_with_arg_passed_to_Block(self):
        bs = BlockState({"snowy": True})
        b = Block("dirt", blockstate=bs)

        diff = str(b.blockstate)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_with_args_passed_to_Block(self):
        bs = BlockState({"snowy": True, "facing": "north"})
        b = Block("dirt", blockstate=bs)

        diff = str(b.blockstate)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attr_through_Block(self):
        b = Block("dirt")
        b.blockstate.snowy = True

        diff = str(b.blockstate)
        test = "[snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_via_attrs_through_Block(self):
        b = Block("dirt")
        b.blockstate.snowy = True
        b.blockstate.facing = "north"

        diff = str(b.blockstate)
        test = "[facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_mixed_attrs_and_args_through_Block(self):
        bs = BlockState({"snowy": True, "dancing": "well"})
        b = Block("dirt")
        b.blockstate = bs
        b.blockstate.facing = "north"

        diff = str(b.blockstate)
        test = "[dancing=well,facing=north,snowy=true]"
        self.assertEqual(diff, test)

    def test_blockstate_override_attr_through_Block(self):
        bs = BlockState({"snowy": True})
        b = Block("dirt")
        b.blockstate = bs
        b.blockstate.snowy = False

        diff = str(b.blockstate)
        test = "[snowy=false]"
        self.assertEqual(diff, test)

    def test_blockstate_incorrect_argument_value_type(self):
        bs = BlockState({"snowy": ["test", "test1"]})

        with self.assertRaises(UnexpectedBlockStatePropertyValueType):
            str(bs)



from .BlockState import BlockState
from .NBT import NBT


class Block:
    """
    A single '<block>' argument looks like this:

        'stone'
        'minecraft:redstone_wire[power=15,north=up,south=side]'
        'minecraft:jukebox{RecordItem:{...}}'
        'minecraft:furnace[facing=north]{BurnTime:200}'

    The format of '<block>' parameters is 'namespaced_ID[block_states]{data_tags}', in which
    block states and data tags can be omitted when they are not needed.

        Namespaced ID is required (though if namespace isn't set it defaults to 'minecraft:').
            In the context of "conditions"/testing for blocks, it can also be the namespace ID
            of block tag with the prefix of '#', such as '#minecraft:planks'.

        Block states are inside '[]', comma-separated and must be properties/values supported
        by the blocks. They are optional.
            'minecraft:stone[doesntexist=purpleberry]' is a syntax error, because 'stone'
            doesn't have 'doesntexist'.
            'minecraft:redstone_wire[power=tuesday]' is a syntax error, because 'redstone_wire''s
            'power' is a number between 0 and 15.

        Data tags are inside '{}'. It's optional.

        In the context of "conditions"/testing for blocks, only the states provided are tested.
            If command tests 'redstone_wire[power=15]', it checks only power, but ignores other
            states such as north.

        In the context of setting blocks, any states provided are set, but anything omitted retain
        their default values, depending on the block.
            If command sets 'redstone_wire[power=15]', it is set 'power' to 15, but 'north' is a
            default value (in this case, set to 'none').
    """

    def __init__(self, id: str, blockstate: BlockState = None, nbt: NBT = None):

        if not isinstance(id, str) or id == "":
            raise ValueError("id must be a non empty string")

        if ":" in id:
            parsed = id.split(":")
            self.namespace = parsed[0]
            self.id = parsed[1]
        else:
            self.namespace = "minecraft"
            self.id = id

        self.blockstate = (
            blockstate if isinstance(blockstate, BlockState) else BlockState()
        )
        self.nbt = nbt if isinstance(nbt, NBT) else NBT()

    def __str__(self):
        return f"{self.namespace}:{self.id}{self.blockstate}{self.nbt}"



import unittest

from shulker.components.Block import Block
from shulker.components.Coordinates import Coordinates
from shulker.components.BlockHandler import BlockHandler
from shulker.components.BlockState import BlockState


class TestBlockHandler(unittest.TestCase):
    def test_handler_real_case(self):
        command = f'setblock {Coordinates(0, 0, 0)} {Block("bedrock")} {BlockHandler("destroy")}'

        diff = str(command)
        test = "setblock 0 0 0 minecraft:bedrock destroy"
        self.assertEqual(diff, test)

    def test_fill_with_blockstate(self):
        pos1 = Coordinates(0, 4, 0)
        pos2 = Coordinates(10, 4, 10)
        bs = BlockState({"facing": "north", "half": "top"})
        block = Block("oak_stairs", blockstate=bs)

        command = f"fill {pos1} {pos2} {block}"

        diff = str(command)
        expected = "fill 0 4 0 10 4 10 minecraft:oak_stairs[facing=north,half=top]"
        self.assertEqual(diff, expected)



from typing import Union

from shulker.components.Block import Block
from shulker.components.BlockState import BlockState
from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.BlockHandler import BlockHandler

from shulker.functions.base_functions import *


def meta_get_block(coords: BlockCoordinates, block: Block, handler: BlockHandler) -> dict:
    return f"setblock {coords} {block} {handler}"


def get_block(
    coords: Union[BlockCoordinates, tuple],
) -> Block:

    """
    Returns a Block
    Available handlers:
        'replace' — The old block drops neither itself nor any contents. Plays no sound.
        'destroy' — The old block drops both itself and its contents (as if destroyed by a player). Plays the appropriate block breaking noise.
        'keep' — Only air blocks are changed (non-air blocks are unchanged).

    Defaults to 'replace'
    """

    check_output_channel()

    coords = format_arg(coords, BlockCoordinates)

    instructions = meta_get_block(coords)

    for line in instructions["list"]:
        post(line)

    return


meta_definition = "custom"

"""
1. Check if block above is air/or +256
2. If not -> store the block type
3. Spawn the arrow as close as needed
3bis. If the arrow didn't hit anything, bruteforce check the block, otherwise:
4. Check the arrow inBlockState nbt data
5. Put the block back
"""



import signal
import sys

from .singleton import singleton
from mctools import RCONClient


class RconClient:
    def __init__(self, ip, password, port=25575):
        self.ip = ip
        self.password = password
        self.port = port
        self.mcr = RCONClient(self.ip, port=self.port)

        try:
            self.mcr.login(self.password)
        except Exception as e:
            raise e

        singleton.add_output_channel(self.mcr.command)



import logging

from mcstatus import JavaServer

logger = logging.getLogger()
logger.setLevel(logging.INFO)

def ping(ip, port):
    server = JavaServer(ip, port)
    try:
        status = server.status()
    except:
        raise Exception
    return status.latency


from typing import Union

from shulker.components.Block import Block
from shulker.components.BlockState import BlockState
from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.BlockHandler import BlockHandler

from shulker.functions.base_functions import *


def meta_set_block(coords: BlockCoordinates, block: Block, handler: BlockHandler) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"setblock {coords} {block} {handler}")
    return instructions


def set_block(
    coords: Union[BlockCoordinates, tuple],
    block: Union[Block, str],
    handler: Union[BlockHandler, str] = "replace",
) -> Union[bool, str]:
    """
    Returns a bool that is set to True
    if no message was sent back by the game or the
    message itself if there was an issue

    Available handlers:
        'replace' — The old block drops neither itself nor any contents. Plays no sound.
        'destroy' — The old block drops both itself and its contents (as if destroyed by a player). Plays the appropriate block breaking noise.
        'keep' — Only air blocks are changed (non-air blocks are unchanged).

    Defaults to 'replace'
    """

    check_output_channel()

    coords = format_arg(coords, BlockCoordinates)
    block = format_arg(block, Block)
    handler = format_arg(handler, BlockHandler)

    instructions = meta_set_block(coords, block, handler)

    for line in instructions["list"]:
        status = post(line)

    return True if status.startswith("Changed the block") else status


meta_definition = {
    "setblock": {
        "type": "literal",
        "children": {
            "pos": {
                "type": "argument",
                "parser": BlockCoordinates,
                "children": {
                    "block": {
                        "type": "argument",
                        "parser": BlockState,
                        "children": {
                            "destroy": {"type": "literal", "executable": True},
                            "keep": {"type": "literal", "executable": True},
                            "replace": {"type": "literal", "executable": True},
                        },
                        "executable": True,
                    }
                },
            }
        },
    }
}



import unittest

from shulker.components.Block import Block
from shulker.components.BlockState import BlockState
from shulker.components.NBT import NBT


class TestBlock(unittest.TestCase):
    def test_block_bedrock(self):
        block = Block("bedrock")

        diff = str(block)
        test = "minecraft:bedrock"
        self.assertEqual(diff, test)

    def test_block_namespace(self):
        block = Block("mod:op_block")

        diff = str(block)
        test = "mod:op_block"
        self.assertEqual(diff, test)

    def test_block_simple_blockstate(self):
        block = Block("bedrock")
        block.blockstate = BlockState({"facing": "north"})

        diff = str(block)
        test = "minecraft:bedrock[facing=north]"
        self.assertEqual(diff, test)

    def test_block_double_blockstate(self):
        block = Block("bedrock")
        block.blockstate = BlockState({"facing": "north", "half": "top"})

        diff = str(block)
        test = "minecraft:bedrock[facing=north,half=top]"
        self.assertEqual(diff, test)

    def test_block_simple_nbt(self):
        block = Block("bedrock")
        block.nbt = NBT({"Fire": 4})

        diff = str(block)
        test = "minecraft:bedrock{Fire:4}"
        self.assertEqual(diff, test)

    def test_block_double_nbt(self):
        block = Block("bedrock")
        block.nbt = NBT({"Fire": 4, "Air": 0})

        diff = str(block)
        test = "minecraft:bedrock{Air:0,Fire:4}"
        self.assertEqual(diff, test)

    def test_block_mixed(self):
        block = Block("bedrock")
        block.nbt = NBT({"Fire": 4, "Air": 0})
        block.blockstate = BlockState({"facing": "north", "half": "top"})

        diff = str(block)
        test = "minecraft:bedrock[facing=north,half=top]{Air:0,Fire:4}"
        self.assertEqual(diff, test)

    def test_empty(self):
        with self.assertRaises(ValueError):
            Block("")


if __name__ == "__main__":
    unittest.main()



class BlockHandler:
    """
    DESTROY: Replaces all blocks (including air) in the fill region with the specified block,
    dropping the existing blocks (including those that are unchanged) and block contents
    as entities as if they had been mined with an unenchanted diamond shovel or pickaxe.
    (Blocks that can only be mined with shears, such as vines, will not drop; neither will liquids.)

    HOLLOW: Replaces only blocks on the outer edge of the fill region with the specified block.
    Inner blocks are changed to air, dropping their contents as entities but not themselves.
    If the fill region has no inner blocks (because it is smaller than three blocks
    in at least one dimension), acts like replace.

    KEEP: Replaces only air blocks in the fill region with the specified block.

    OUTLINE: Replaces only blocks on the outer edge of the fill region with the specified block.
    Inner blocks are not affected. If the fill region has no inner blocks
    (because it is smaller than three blocks in at least one dimension), acts like replace.

    REPLACE: Replaces all blocks (including air) in the fill region with the specified block,
    without dropping blocks or block contents as entities. Optionally, instead of specifying
    a data tag for the replacing block, block ID and data values may be specified to limit
    which blocks are replaced (see replaceTileName and replaceDataValue below)
    """

    options = ["destroy", "hollow", "keep", "outline", "replace"]

    def __init__(self, option: str = "replace"):
        self.option = option

    def __str__(self):
        if self.option not in self.options:
            raise BlockHandlerWrongType(
                f'The BlockHandler provided: \'{self.option}\' is not a valid option. Availables: [{" | ".join(self.options)}]'
            )
        else:
            return f"{self.option}"


class BlockHandlerWrongType(Exception):
    pass


class BlockHandlerWrongOption(Exception):
    pass



from typing import Union
from shulker.functions.base_functions import *


def meta_say(text: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"say {text}")
    return instructions


def say(
    text: str,
) -> Union[bool, str]:
    """
    Returns a bool that is set to True
    if no message was sent back by the game or the
    message itself if there was an issue

    This function sends a message in the chat
    """

    check_output_channel()
    
    if type(text) is not str:
        raise TypeError(f"Expected type str, got {type(text)}")

    instructions = meta_say(text)

    for line in instructions["list"]:
        status = post(line)

    return True if status.startswith("") else status



from .Block import Block
from .BlockCoordinates import BlockCoordinates
from .BlockHandler import BlockHandler
from .BlockState import BlockState
from .Coordinates import Coordinates
from .Entity import Entity
from .NBT import NBT
from .TargetSelector import TargetSelector

from .Zone import Zone


from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.server.singleton import singleton

__all__ = [
    "check_output_channel",
    "post",
    "default_check",
    "execute_check",
    "unexpected_status",
    "format_arg",
    "nest_commands",
    "post_nest",
]


def check_output_channel():
    if singleton.check_status():
        return singleton
    else:
        raise NoOutputChannelProvided(f"No output channel was initialized in the code")


def post(cmd: str):
    return singleton.post(cmd)


def post_nest(coords: BlockCoordinates, nest: list):
    for index, egg in enumerate(nest):
        cmd = f'/setblock {coords.x} {coords.y + index + 2} {coords.z} command_block{{Command:"{egg}",auto:1b,conditionMet:1b}} replace'
        cmd += "\n"
        print(cmd)
        singleton.post(cmd)


def nesting_process(instructions):

    nested_cmd = f"summon falling_block ~ ~1 ~"

    nested_cmd += " {Time:1,BlockState:{Name:redstone_block},Passengers:[{id:falling_block,Passengers:[{id:falling_block,Time:1,BlockState:{Name:activator_rail},Passengers:["

    for line in instructions:
        nested_cmd += f'{{id:command_block_minecart,Command:"{line}"}},'

    nested_cmd += "{id:command_block_minecart,Command:'setblock ~ ~1 ~ command_block{auto:1,Command:\"fill ~ ~ ~ ~ ~-3 ~ air\"}'},"
    nested_cmd += "{id:command_block_minecart,Command:'kill @e[distance=..1]'}]}]}]}"

    return nested_cmd


def nest_commands(instructions):

    nest = []

    pckg = []
    total_length = 0
    while len(instructions):
        print(len(instructions))
        if total_length < 1600:
            total_length += len(instructions[0])
            pckg.append(instructions[0])
            instructions.pop(0)
        else:
            nest.append(nesting_process(pckg))
            pckg = []
            total_length = 0

    return nest


def default_check(response):
    if response == "":
        return True
    else:
        return response


def execute_check(response):
    if response == "Test passed":
        return True
    elif response == "Test failed":
        return False
    else:
        return response


def unexpected_status(file_name, status):
    raise UnexpectedReturn(
        f'The say command in {file_name.split("/")[-1]} didn\'t properly function and returned: "{status}"'
    )


def format_arg(argument, component):
    if isinstance(argument, component):
        return argument

    if isinstance(argument, str):
        return component(argument)
    elif isinstance(argument, tuple):
        return component(*argument)
    elif isinstance(argument, list):
        return component(*argument)
    else:
        raise InvalidArgumentType(
            f'Invalid argument type, could not parse "{argument}" as a valid component.'
        )


class InvalidArgumentType(Exception):
    pass


class UnexpectedReturn(Exception):
    pass


class NoOutputChannelProvided(Exception):
    pass



import unittest

from shulker.components.Coordinates import WrongCaretNotation
from shulker.components.BlockCoordinates import BlockCoordinates


class TestCoordinates(unittest.TestCase):
    def test_blockcoords_zeros(self):
        coords = BlockCoordinates(0, 0, 0)

        diff = str(coords)
        test = "0 0 0"
        self.assertEqual(diff, test)

    def test_coords_tilde(self):
        coords = BlockCoordinates("~", "~", "~")

        diff = str(coords)
        test = "~ ~ ~"
        self.assertEqual(diff, test)

    def test_coords_tilde_mixed(self):
        coords = BlockCoordinates("~5", 0, "~-10")

        diff = str(coords)
        test = "~5 0 ~-10"
        self.assertEqual(diff, test)

    def test_coords_caret(self):
        coords = BlockCoordinates("^", "^", "^")

        diff = str(coords)
        test = "^ ^ ^"
        self.assertEqual(diff, test)

    def test_coords_caret_mixed(self):
        coords = BlockCoordinates("^5", "^-10", "^")

        diff = str(coords)
        test = "^5 ^-10 ^"
        self.assertEqual(diff, test)

    def test_coords_caret_wrong(self):
        coords = BlockCoordinates(5, "^", "^")

        with self.assertRaises(WrongCaretNotation):
            str(coords)

    def test_coords_floats(self):
        coords = BlockCoordinates(1.5, "~0.1", 1)

        diff = str(coords)
        test = "1 ~0 1"
        self.assertEqual(diff, test)


if __name__ == "__main__":
    unittest.main()



import json

from typing import Union

from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.NBT import NBT

from shulker.functions.base_functions import *


def meta_summon(entity: str, coords: BlockCoordinates, nbt_data: NBT) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"summon {entity} {coords} {nbt_data}")
    return instructions


def summon(
    entity: str,
    coords: Union[BlockCoordinates, tuple],
    nbt_data: Union[NBT, dict, str]
) -> Union[bool, str]:
    """
    Returns a bool that is set to True
    if no message was sent back by the game or the
    message itself if there was an issue

    Summons an entity
    """

    check_output_channel()
    
    # Get the list of entity from registries.json
    with open("mc_data/registries.json", "r") as f:
        registries = json.load(f)
        entities = registries["minecraft:entity_type"]["entries"]
        
    if type(entity) is not str:
        raise TypeError(f"Expected type str, got {type(entity)}")
    elif entity.replace("minecraft:", "") not in entities.replace("minecraft:", ""):
        raise ValueError(f"Entity {entity} is not a valid entity")
    
    coords = format_arg(coords, BlockCoordinates)
    nbt_data = format_arg(nbt_data, NBT)

    instructions = meta_summon(entity, coords, nbt_data)

    for line in instructions["list"]:
        status = post(line)

    return True if status.startswith("") else status



import math

from .Coordinates import Coordinates


class BlockCoordinates(Coordinates):
    """
    The position of a block is actually the coordinates of the point at the lower northwest corner
    of the block, that is, the integer coordinates obtained by rounding down the coordinates
    inside the block.

    In Minecraft, decimal coordinates usually needs to be converted into integer coordinates by rounding
    down, which is called the block position of the coordinate.
    """

    def floor(self, coord):
        # Below is the worst bricolage in the history of bricolage
        # '^1.1' becomes '^1'

        if isinstance(coord, str) and "." in coord:

            if coord.startswith("^"):
                value = coord.split("^")[1]
                value = float(value)
                value = math.floor(value)
                coord = "^" + str(value)

            elif coord.startswith("~"):
                value = coord.split("~")[1]
                value = float(value)
                value = math.floor(value)
                coord = "~" + str(value)

            else:
                coord = math.floor(float(coord))

        elif isinstance(coord, float):
            coord = math.floor(coord)

        return coord

    def __str__(self):

        self.check_carets()

        self.x = self.floor(self.x)
        self.y = self.floor(self.y)
        self.z = self.floor(self.z)

        return f"{self.x} {self.y} {self.z}"



from .base_functions import nest_commands
from .get_block import get_block, meta_get_block
from .set_block import set_block, meta_set_block
from .set_image import set_image, meta_set_image, print_palette
from .set_text import set_text, meta_set_text
from .set_zone import set_zone, meta_set_zone
from .say import say, meta_say
from .summon import summon, meta_summon
from .get_player_info import get_player_info, get_player_pos, meta_get_player_info


import re
from typing import Union

from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.Coordinates import Coordinates
from shulker.components.NBT import NBT

from shulker.functions.base_functions import *


def meta_get_player_info(pseudo: str) -> dict:
    instructions = {"list": []}
    instructions["list"].append(f"data get entity {pseudo}")
    return instructions


def get_player_info(
    pseudo: str,
) -> Union[bool, str]:
    """
    Returns the NBT data of the player
    """

    check_output_channel()
    
    if type(pseudo) is not str:
        raise TypeError(f"Expected type str, got {type(pseudo)}")

    instructions = meta_get_player_info(pseudo)

    for line in instructions["list"]:
        status = post(line)
        
    if status.startswith(f"{pseudo} has"):
        return NBT(status.split("data: ")[1])
    else:
        return False
    

def get_player_pos(
    pseudo: str,
    rounded: bool = True
) -> Union[bool, str]:
    """
    Returns the coordinates of the player if it is found
    False if it wasn't found or there's an issue
    """

    check_output_channel()
    
    if type(pseudo) is not str:
        raise TypeError(f"Expected type str, got {type(pseudo)}")

    instructions = meta_get_player_info(pseudo)

    for line in instructions["list"]:
        status = post(line)
        
    if status.startswith(f"{pseudo} has"):
        matches = re.findall(r"Pos: \[(.+?)\..+?, (.+?)\..+?, (.+?)\..+?\]", status)
        if len(matches) == 3:
            if rounded:
                return BlockCoordinates(*tuple(map(round, matches)))
            else:
                return Coordinates(*tuple(matches))
    else:
        return False



import urllib3
import json
import math
from typing import Union
import sys
import os

from PIL import Image

from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.BlockHandler import BlockHandler
from shulker.components.Zone import Zone

from shulker.functions.base_functions import *
from shulker.functions.set_block import meta_set_block, set_block

# TODO: Handle image in memory rather than pull
# TODO: Minecart method? -> https://www.youtube.com/watch?v=MEawKJm-t28


def meta_set_image(
    file: str, coords: BlockCoordinates, orientation: str, player_name: str
) -> dict:

    image = Image.open(file)
    converted_image = image.convert("RGB")

    pixels = list(converted_image.getdata())
    width, height = converted_image.size
    pixels = [pixels[i * width : (i + 1) * width] for i in range(height)]

    instructions = {"list": [], "zone": None}

    palette = get_palette(orientation)

    x = 0

    for line in pixels:
        x += 1
        z = 0

        for pixel in line:

            block = color_picker(pixel, palette)

            if player_name is not None:
                # TODO: Update with the TP method later on
                cmd = f"tp {player_name} {x} {coords.y + 100} {z}"
                instructions["list"].append(cmd)

            new_coords = BlockCoordinates(coords.x + x, coords.y, coords.z + z)

            if orientation in block:
                block = block.replace(f"_{orientation}", "")

                if orientation == "side":
                    block.blockhandler.axis = "x"

                elif orientation == "top":
                    block.blockhandler.axis = "x"

            cmd = meta_set_block(
                new_coords,
                block,
                BlockHandler("replace"),
            )
            instructions["list"].append(cmd)

            z += 1
    else:
        instructions["zone"] = Zone(
            coords, BlockCoordinates(coords.x + x, coords.y, coords.z + z)
        )

    return instructions


def set_image(
    file: str,
    coords: Union[BlockCoordinates, tuple],
    orientation: str = "side",
    url: bool = False,
    player_name: str = None,
) -> Zone:
    """
    This function takes the path to an image or an URL
    (in which case, url must be passed as True), downloads it,
    and sends back the setblocks instruction to print the image
    in the provided orientation

    Orientation can either be "side", "top", "bottom" depending from where
    you want people to look at the image.

    If a player name is provided, the player will be teleported to make sure the
    printing doesn't happen out of bound (and therefore fail)
    """

    check_output_channel()

    # TODO: Cleanup the file
    if url == True:
        http = urllib3.PoolManager()
        r = http.request("GET", file, preload_content=False)

        filename = "image.jpg"
        with open(filename, "wb") as out:
            while True:
                data = r.read(1024)
                if not data:
                    break
                out.write(data)

        r.release_conn()
        file = filename

    coords = format_arg(coords, BlockCoordinates)

    if orientation not in ["side", "top", "bottom"]:
        raise ValueError(
            f"Orientation must either be side, top or bottom for set_image()"
        )

    instructions = meta_set_image(file, coords, orientation, player_name)

    for line in instructions["list"]:
        post(line)

    """nest = nest_commands(instructions["lines"])
    post_nest(coords, nest)"""

    return instructions["zone"]


def get_palette(orientation):

    try:
        path = os.path.dirname(os.path.abspath(__file__))
        with open(f"{path}/mc_data/palette.json") as f:
            palette = json.load(f)

    except FileNotFoundError:
        palette = create_palette()

    return palette[orientation]


def get_closest_color(pixel, block_palette, orientation):

    path = os.path.dirname(os.path.abspath(__file__))
    with open(f"{path}/mc_data/block_list.json") as f:
        block_list = json.load(f)

    orientations = ["top", "bottom", "front"]
    for index, value in enumerate(orientations):
        if value == orientation:
            orientations.pop(index)

    excluded_block_tags = [
        "shulker",
        "sand",
        "powder",
        "glass",
        "front",
        "coral",
        "dead",
        "trapdoor",
        "chorus",
        "_ore",
    ]

    excluded_blocks = [
        "ice",
    ]

    best_delta = None

    for block in block_palette["averages"]:

        block_name = block["texture"].split("/")[-1]

        if any(other_orientation in block_name for other_orientation in orientations):
            continue

        if f"minecraft:{block_name}" not in list(block_list.keys()):
            continue

        elif any(to_exclude in block_name for to_exclude in excluded_block_tags):
            continue

        elif any(to_exclude == block_name for to_exclude in excluded_blocks):
            continue

        elif block["pixels"] != 256:
            continue

        elif block["stddev_rgb"] == 0.00000:
            continue

        delta = math.sqrt(
            (pixel[0] - block["rgba"][0]) ** 2
            + (pixel[1] - block["rgba"][1]) ** 2
            + (pixel[2] - block["rgba"][2]) ** 2
        )

        if best_delta is None or delta < best_delta:
            best_delta = delta
            best_match = block_name

    return best_match


def color_picker(pixel, palette):

    r = int(pixel[0] / 32)
    g = int(pixel[1] / 32)
    b = int(pixel[2] / 32)

    index = r * 64 + g * 8 + b
    return palette[str(index)]


def print_palette(coords, orientation="side"):
    palette = get_palette(orientation)

    n = 8
    for x in range(n):
        for y in range(n):
            for z in range(n):
                coords = BlockCoordinates(x * 3, y * 3 + 5, z * 3)

                set_block(
                    coords,
                    palette[str(x * 64 + y * 8 + z)],
                )


def create_palette():

    print(
        "It looks like this is the first time you're using set_image() and no pre-existing block palette was found. Please hold on while it's generating!"
    )

    path = os.path.dirname(os.path.abspath(__file__))
    with open(f"{path}/mc_data/blocks_color.json") as f:
        block_palette = json.load(f)

    palette = {"side": {}, "top": {}, "bottom": {}}

    for orientation in ["side", "top", "bottom"]:

        print(f"-> Generating the palette for the '{orientation}' orientation")

        for i in range(0, 8):
            for j in range(0, 8):
                for k in range(0, 8):

                    block = get_closest_color(
                        (i * 32, j * 32, k * 32), block_palette, orientation
                    )

                    palette[orientation][str(i * 64 + j * 8 + k)] = block

    with open(f"{path}/mc_data/palette.json", "w+") as f:
        json.dump(palette, f, indent=4)

    return palette



class TargetSelector:
    """
    In most commands where entities may be specified as an argument,
    it is possible to "target" one or more entities satisfying
    certain conditions. To target entities by condition,
    choose a target selector variable and, optionally,
    one or more target selector arguments to modify the conditions
    to be satisfied.

    Available identifiers:
        - p	    |= nearest player
        - a	    |= all players
        - r	    |= random player
        - s	    |= entity executing the command
        - e	    |= all entities

    Summary of target selector arguments:

        Selection by Position:
            Argument(s)	 |  Selection criteria   |  Example value
            _____________________________________________________

            x, y, z	     |  coordinate           |  (float, int) 7, +12.34, -0.1
            distance‌     |  distance	          |  (int, str) 8, 16, "8..16" (from 8 to 16)
            dx, dy, dz	 |  volume dimensions    |  (int) 1, 2, 0

        Selection by Scoreboard Values:
            Argument(s)	 |  Selection criteria   |  Example value
            _____________________________________________________

            scores	     |  scores               |  (dict) {myscore=10}, {foo=..15}
            *tag	     |  tag                  |  (str) has_tag, !hasnt_tag, "" (=has no tags)
            team	     |  team name            |  (str) in_team, !not_in_team, "" (=has no team)

        Selection by Traits:
            Argument(s)	 |  Selection criteria   |  Example value
            _____________________________________________________

            limit,sort‌   |  limit                |  (str and int) sort=(nearest|furthest|random|arbitrary) limit= (int)
            level‌        |  experience level     |  (int, str) 8, 16, "8..16" (from 8 to 16)
            gamemode‌     |  game mode            |  (str) gamemode=(spectator‌|adventure|creative|survival), !survival
            name	     |  entity name          |  (str) Johnson, !Johnson
            x_rotation   |  vertical rotation    |  (int, str) 8, 16, "8..16" (from 8 to 16)
            y_rotation   |  horizontal rotation  |  (int, str) 8, 16, "8..16" (from 8 to 16)
            type	     |  entity type          |  (str) pig, mod:modded_mob, !pig
            nbt‌          |  nbt                  |  (dict, object) {OnGround:true}, [class] NBT()
            advancements‌ |  advancements         |  (dict) {story/follow_ender_eye=true}
            predicate    |  predicate            |  (str) example:test_predicate

    *can be a list of the value

    You can read in-depth about target selectors at:
        https://minecraft.fandom.com/wiki/Commands#Target_selectors
    """

    argument_list = [
        "x",
        "y",
        "z",
        "dx",
        "dy",
        "dz",
        "distance",
        "scores",
        "tag",
        "team",
        "limit",
        "sort‌",
        "level‌",
        "gamemode‌",
        "name",
        "x_rotation",
        "y_rotation",
        "type",
        "nbt‌",
        "advancements‌",
        "predicate",
    ]

    def __init__(self, identifier: str, arguments: dict = None):

        if identifier not in ["p", "a", "r", "s", "e"]:
            raise IncorrectTargetSelectorIdentifier(
                f"The TargetSelector identifier must be either P, A, R, S or E"
            )

        self.id = identifier

        if not isinstance(arguments, dict) and arguments is not None:
            raise IncorrectTargetSelectorArgumentsType(
                f"The TargetSelector arguments must be passed as a dictionnary"
            )

        self.arguments = arguments

    def __str__(self):
        buff = ""

        if self.arguments is not None:

            for key in self.arguments:

                if key not in self.argument_list:
                    raise InvalidTargetSelectorArgumentKey(
                        f'The key "{key}" is not a valid argument type for a selector'
                    )

                if isinstance(self.arguments[key], list):
                    for element in self.arguments[key]:
                        buff += f"{key}={element},"
                else:
                    buff += f"{key}={self.arguments[key]},"

            if buff != "":
                buff = buff[:-1]

        return f"@{self.id}[{buff}]"


class IncorrectTargetSelectorIdentifier(Exception):
    pass


class IncorrectTargetSelectorArgumentsType(Exception):
    pass


class InvalidTargetSelectorArgumentKey(Exception):
    pass



class OutputChannel():
    def __init__(self):
        self.ready = False
    
    def check_status(self):
        if 'output_channel' in dir(self):
            return True
        else:
            return False
    
    def add_output_channel(self, output_channel):
        self.output_channel = output_channel

    def post(self, cmd):
        if hasattr(self, 'output_channel') and callable(self.output_channel):
            return self.output_channel(cmd)
        else:
            raise NoOutputChannelProvided(f'ServerInstance did not received a correct output channel. Add it with add_output_channel')
 
singleton = OutputChannel()

class NoOutputChannelProvided(Exception):
    pass


import unittest

from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.Zone import Zone, ZoneWrongCoordsType


class TestCoordinates(unittest.TestCase):
    def test_zone_with_blockcoords(self):
        coords1 = BlockCoordinates(0, 0, 0)
        coords2 = BlockCoordinates(10, 10, 10)
        zone = Zone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_with_tuple(self):
        coords1 = (0, 0, 0)
        coords2 = (10, 10, 10)
        zone = Zone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_mixed(self):
        coords1 = (0, 0, 0)
        coords2 = BlockCoordinates(10, 10, 10)
        zone = Zone(coords1, coords2)

        diff = str(zone)
        test = "0 0 0 10 10 10"
        self.assertEqual(diff, test)

    def test_zone_wrong_args(self):
        zone = Zone("hello", 1)
        with self.assertRaises(ZoneWrongCoordsType):
            str(zone)


if __name__ == "__main__":
    unittest.main()



import unittest

from shulker.components.NBT import NBT
from shulker.components.Block import Block


class TestBlock(unittest.TestCase):
    def test_NBT_without_args(self):
        tags = NBT()

        diff = str(tags)
        test = ""
        self.assertEqual(diff, test)

    def test_NBT_without_attrs(self):
        b = Block("dirt")

        diff = str(b.nbt)
        test = ""
        self.assertEqual(diff, test)

    def test_NBT_without_args_passed_to_Block(self):
        tags = NBT()
        b = Block("dirt", nbt=tags)

        diff = str(b.nbt)
        test = ""
        self.assertEqual(diff, test)

    def test_NBT_without_args_attributed_to_Block(self):
        tags = NBT()
        b = Block("dirt")

        b.nbt = tags

        diff = str(b.nbt)
        test = ""
        self.assertEqual(diff, test)

    def test_NBT_with_arg(self):
        tags = NBT({"Lock": "45"})

        diff = str(tags)
        test = '{Lock:"45"}'
        self.assertEqual(diff, test)

    def test_NBT_with_args(self):
        tags = NBT(
            {
                "Fire": 4,
                "Air": 10,
                "AbsorptionAmount": 150.0,
                "Motion": [2.0, 2.0, 2.0],
                "Passengers": [{"id": "minecraft:area_effect_cloud"}],
                "CustomName": '{"text":"test"}',
            }
        )

        diff = str(tags)
        test = "{AbsorptionAmount:150.0d,Air:10,CustomName:'{\"text\":\"test\"}',Fire:4,Motion:[2.0, 2.0, 2.0],Passengers:[{'id': 'minecraft:area_effect_cloud'}]}"
        self.assertEqual(diff, test)

    def test_NBT_via_attr(self):
        tags = NBT()
        tags.Fire = 4

        diff = str(tags)
        test = "{Fire:4}"
        self.assertEqual(diff, test)

    def test_NBT_via_attrs(self):
        tags = NBT()
        tags.Motion = [2.0, 2.0, 2.0]
        tags.Passengers = [{"id": "minecraft:area_effect_cloud"}]

        diff = str(tags)
        test = "{Motion:[2.0, 2.0, 2.0],Passengers:[{'id': 'minecraft:area_effect_cloud'}]}"
        self.assertEqual(diff, test)

    def test_NBT_with_arg_passed_to_Block(self):
        tags = NBT({"Fire": 4})
        b = Block("dirt", nbt=tags)

        diff = str(b.nbt)
        test = "{Fire:4}"
        self.assertEqual(diff, test)

    def test_NBT_with_args_overriden(self):
        tags = NBT({"Fire": 4})
        tags.Fire = 5

        diff = str(tags)
        test = "{Fire:5}"
        self.assertEqual(diff, test)

    def test_NBT_with_args_passed_to_Block(self):
        tags = NBT(
            {
                "Fire": 4,
                "Air": 10,
                "AbsorptionAmount": 150.0,
                "Motion": [2.0, 2.0, 2.0],
                "Passengers": [{"id": "minecraft:area_effect_cloud"}],
                "CustomName": '{"text":"test"}',
            }
        )
        b = Block("dirt", nbt=tags)

        diff = str(b.nbt)
        test = "{AbsorptionAmount:150.0d,Air:10,CustomName:'{\"text\":\"test\"}',Fire:4,Motion:[2.0, 2.0, 2.0],Passengers:[{'id': 'minecraft:area_effect_cloud'}]}"
        self.assertEqual(diff, test)

    def test_NBT_via_attr_through_Block(self):
        b = Block("dirt")
        b.nbt.Fire = 4

        diff = str(b.nbt)
        test = "{Fire:4}"
        self.assertEqual(diff, test)

    def test_NBT_via_attrs_through_Block(self):
        b = Block("dirt")
        b.nbt.Motion = [2.0, 2.0, 2.0]
        b.nbt.Passengers = [{"id": "minecraft:area_effect_cloud"}]

        diff = str(b.nbt)
        test = "{Motion:[2.0, 2.0, 2.0],Passengers:[{'id': 'minecraft:area_effect_cloud'}]}"
        self.assertEqual(diff, test)

    def test_NBT_mixed_attrs_and_args_through_Block(self):
        tags = NBT({"Fire": 4, "CustomName": '{"text":"test"}'})
        b = Block("dirt")
        b.nbt = tags
        b.nbt.Motion = [2.0, 2.0, 2.0]
        b.nbt.Passengers = [{"id": "minecraft:area_effect_cloud"}]

        diff = str(b.nbt)
        test = "{CustomName:'{\"text\":\"test\"}',Fire:4,Motion:[2.0, 2.0, 2.0],Passengers:[{'id': 'minecraft:area_effect_cloud'}]}"
        self.assertEqual(diff, test)



from typing import Union

from shulker.components.Block import Block
from shulker.components.BlockCoordinates import BlockCoordinates
from shulker.components.BlockState import BlockState
from shulker.components.BlockHandler import BlockHandler
from shulker.components.Zone import Zone

from shulker.functions.base_functions import *
from shulker.functions.set_block import meta_set_block
from shulker.functions.set_zone import meta_set_zone

# TODO: Make clear zone works with scaling
# TODO: orientation
# TODO: vertical text (like japanese)
# TODO: Remove digit skip and encode them
# TODO: handle '\n'
# TODO: flatten scale for every number to usable (but translates to odd numbers)


def meta_set_text(
    message: str,
    coords: BlockCoordinates,
    palette: list,
    style: str,
    orientation: str,
    replace: BlockHandler,
    scale: int,
    filler: Block,
) -> dict:

    handler = BlockHandler("replace" if replace else "keep")

    offset = 0
    instructions = {"list": [], "zone": None}

    # Can't use enumerate(message) because I don't want
    # to count spaces
    letter_index = 0

    # First we iterate over each letter of the string that
    # was passed
    for index, letter in enumerate(message):

        # If the letter is a space, we skip the loop
        # but add an offset of 2 to the next letter placement
        if letter == " " or letter == "\n":
            offset += 2 + scale
            continue

        try:
            # This condition makes the style "mixed"
            # set the first letter of a string as an uppercase
            if style == "mixed":

                if letter.isupper():
                    letter = letter.lower()
                    representation = letters_dict[letter]["uppercase"]
                else:
                    representation = letters_dict[letter]["lowercase"]

            # Otherwise, it fetches the "representation" (e.g.: 000-421-etc.)
            # of the current letter in the string.
            # The representations are at the bottom of this file.
            else:
                letter = letter.lower()
                representation = letters_dict[letter][style]
        except KeyError:
            offset += 2 + scale
            print("Couldn't find representation for " + letter)
            continue

        # Here we reverse the representation to read it
        # from the bottom left corner to the top right corner
        representation = representation.split("-")
        representation.reverse()

        # This handles when multiple palettes were provided as a list
        # using a modulo based on the length of the palette.
        if len(palette) > 1:
            mod = letter_index % len(palette)
        else:
            mod = 0

        # Then it starts looping through each digit in
        # the representation (e.g.: ["000"], ["414"], etc.)
        y = 0
        for line in representation:

            # After each line it resets the X position to the left
            x = 0 + offset

            for digit in line:

                # This variable stores the computed coordinates at
                # which the current block should be placed
                new_coords = BlockCoordinates(coords.x + x, coords.y + y, coords.z)

                # After which we move the cursor
                x += 1 + scale

                # And finally, this is the part where the block is set
                # depending on whether a scale was specified or not
                if scale == 0:
                    block = fetch_block(digit, orientation, palette[mod], filler)

                    cmd = meta_set_block(new_coords, block, handler)
                    instructions["list"].append(cmd)

                # In case it was, it creates "zones" instead
                # of placing blocks
                else:
                    zones = construct_zones(new_coords, digit, scale)

                    for zone in zones:

                        # A letter scaled up can be represented by multiple zones
                        # with the slabs for instance, it's half filler(/air),
                        # half solid, so we have to adjust which block
                        # is used per sub zone
                        if zone["filler"] == "air":
                            block = filler
                        elif zone["filler"] == "block":
                            temp_digit = "0" if digit == "0" else "1"
                            block = fetch_block(
                                temp_digit, orientation, palette[mod], filler
                            )

                        cmd = meta_set_zone(zone["coords"], block, handler)
                        instructions["list"].append(cmd)

            # After a line has been printed, the Y level is increased
            y += 1 + scale

        # This increases the offset after a letter has been printed
        # by computing the length of any 'line' representing
        # the current letter, and adding one block space
        offset += len(representation[0]) * (scale + 1) + 1 * (scale + 1)
        letter_index += 1

    # Y is being re-declared to be used in the return value
    # of the function
    else:
        try:
            y = representation.count("-")
        except UnboundLocalError:
            y = 0
            new_coords = BlockCoordinates(coords.x, coords.y, coords.z)

    new_coords.y += y
    instructions["zone"] = Zone(coords, new_coords)

    # The functions returns the "instructions" dictionary
    # containing all the commands to .post() as well
    # as the zone representing the text that was created
    # for later clean-up
    return instructions


def set_text(
    message: str,
    coords: Union[BlockCoordinates, tuple],
    palette: Union[list, str] = "quartz",
    style: str = "mixed",
    orientation: str = "east",
    replace: bool = True,
    scale: int = 0,
    filler: Union[Block, str] = "air",
) -> Zone:

    """
    Returns a Zone representing the area used by the text
    to be represented

    The given coords should represent the block that is
    lower northwest corner of the structure

    Available styles:
    - smallcaps
    - lowercase
    - uppercase
    - mixed (alternate between lowercase/uppercase by respecting case)

    The availables palettes are listed at the bottom of this file.

    Orientation can be either east(/west) or north(/south)

    Replace is a boolean which defines whether the 'empty' part of
    the text should be replaced by air blocks or left as is

    Scale is the amount of block to "add" to the size of a default block (1)
    e.g: if scale is set to 1, a block will be represented by a 2x2 zone

    Filler represents the block to use in the empty area of the letter
    representation if Replace is set to True
    """

    check_output_channel()

    coords = format_arg(coords, BlockCoordinates)
    filler = format_arg(filler, Block)

    if type(palette) is not list:
        palette = [palette]

    if scale != 0 and (scale % 2) == 0:
        raise ValueError(f"Scale must be an odd number")

    # TODO: Handle this
    """if orientation == "south":
        orientation = "north"
    elif orientation == "east":
        orientation = "west"
    elif orientation != "west" and orientation != "north":
        raise ValueError(f"Orientation must be either north or west for set_text()")"""

    instructions = meta_set_text(
        message, coords, palette, style, orientation, replace, scale, filler
    )

    for line in instructions["list"]:
        post(line)

    return instructions["zone"]


def construct_zones(coords: BlockCoordinates, digit: str, scale: int) -> list:

    zones = []

    if digit in ["0", "1"]:
        new_coords1 = BlockCoordinates(coords.x, coords.y, coords.z)

        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale, coords.z + scale
        )

        zones.append({"coords": Zone(new_coords1, new_coords2), "filler": "block"})

    elif digit in ["2", "3"]:

        new_coords1 = BlockCoordinates(coords.x, coords.y + ((scale / 2) + 1), coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block" if digit == "2" else "air",
            }
        )

        new_coords1 = BlockCoordinates(coords.x, coords.y, coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale / 2, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block" if digit == "3" else "air",
            }
        )

    elif digit in ["4", "5"]:

        new_coords1 = BlockCoordinates(coords.x, coords.y, coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale / 2, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block",
            }
        )

        new_coords1 = BlockCoordinates(
            coords.x + ((scale / 2) + 1), coords.y + ((scale / 2) + 1), coords.z
        )
        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block" if digit == "4" else "air",
            }
        )

        new_coords1 = BlockCoordinates(coords.x, coords.y + ((scale / 2) + 1), coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale / 2, coords.y + scale, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block" if digit == "5" else "air",
            }
        )

    elif digit in ["6", "7"]:

        new_coords1 = BlockCoordinates(coords.x, coords.y + ((scale / 2) + 1), coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block",
            }
        )

        new_coords1 = BlockCoordinates(coords.x + ((scale / 2) + 1), coords.y, coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale, coords.y + scale / 2, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block" if digit == "6" else "air",
            }
        )

        new_coords1 = BlockCoordinates(coords.x, coords.y, coords.z)
        new_coords2 = BlockCoordinates(
            coords.x + scale / 2, coords.y + scale / 2, coords.z + scale
        )

        zones.append(
            {
                "coords": Zone(new_coords1, new_coords2),
                "filler": "block" if digit == "7" else "air",
            }
        )

    else:
        zones = []

    return zones


def fetch_block(digit: str, orientation: str, palette: str, filler: Block) -> Block:

    counter_orientation = "west" if orientation == "east" else "south"

    if digit == "1":
        block = block_palette[palette]["block"]

    elif digit == "2":
        block = block_palette[palette]["slab"]
        block.blockstate = BlockState({"type": "top"})

    elif digit == "3":
        block = block_palette[palette]["slab"]
        block.blockstate = BlockState({"type": "bottom"})

    elif digit == "4":
        block = block_palette[palette]["stairs"]
        block.blockstate = BlockState(
            {"facing": orientation, "shape": "straight", "half": "bottom"}
        )

    elif digit == "5":
        block = block_palette[palette]["stairs"]
        block.blockstate = BlockState(
            {"facing": counter_orientation, "shape": "straight", "half": "bottom"}
        )

    elif digit == "6":
        block = block_palette[palette]["stairs"]
        block.blockstate = BlockState(
            {"facing": orientation, "shape": "straight", "half": "top"}
        )

    elif digit == "7":
        block = block_palette[palette]["stairs"]
        block.blockstate = BlockState(
            {"facing": counter_orientation, "shape": "straight", "half": "top"}
        )

    else:
        block = filler

    return block


meta_definition = "custom"

""" Legend: 
  0 = void
  1 = block
  2 = up slab
  3 = low slab
 for the 4 stairs below, they are represented by where their empty quarter of block is
  4 = upper left
  5 = upper right
  6 = bottom left
  7 = bottom right
 each letter is represented by a string of 15 digits,
 separated by a dash for each new column,
 representing a letter in a 3*5 block format

 Source: https://cdn.discordapp.com/attachments/414084871120617472/840083817913385000/ccfe08c.png
"""

letters_dict = {
    "a": {
        "smallcaps": "000-425-131-101-000",
        "uppercase": "425-101-121-101-000",
        "lowercase": "000-425-321-636-000",
    },
    "b": {
        "smallcaps": "000-125-125-137-000",
        "uppercase": "125-135-101-137-000",
        "lowercase": "100-125-101-737-000",
    },
    "c": {
        "smallcaps": "000-425-100-637-000",
        "uppercase": "425-100-100-637-000",
        "lowercase": "000-425-100-637-000",
    },
    "d": {
        "smallcaps": "000-125-101-137-000",
        "uppercase": "125-101-101-137-000",
        "lowercase": "001-421-101-636-000",
    },
    "e": {
        "smallcaps": "000-122-120-133-000",
        "uppercase": "122-130-100-133-000",
        "lowercase": "000-425-122-637-000",
    },
    "f": {
        "smallcaps": "000-122-120-100-000",
        "uppercase": "122-130-100-100-000",
        "lowercase": "425-130-100-100-000",
    },
    "g": {
        "smallcaps": "000-425-103-634-000",
        "uppercase": "425-100-106-637-000",
        "lowercase": "000-424-101-021-637",
    },
    "h": {
        "smallcaps": "000-101-121-101-000",
        "uppercase": "101-131-101-101-000",
        "lowercase": "100-125-101-101-000",
    },
    "i": {
        "smallcaps": "000-212-010-313-000",
        "uppercase": "212-010-010-313-000",
        "lowercase": "020-210-010-313-000",
    },
    "j": {
        "smallcaps": "000-001-001-637-000",
        "uppercase": "001-001-001-637-000",
        "lowercase": "002-021-001-001-637",
    },
    "k": {
        "smallcaps": "000-107-163-101-000",
        "uppercase": "107-140-105-101-000",
        "lowercase": "100-103-160-105-000",
    },
    "l": {
        "smallcaps": "000-100-100-133-000",
        "uppercase": "100-100-100-133-000",
        "lowercase": "100-100-100-637-000",
    },
    "m": {
        "smallcaps": "000-504-111-101-000",
        "uppercase": "504-111-101-101-000",
        "lowercase": "000-555-111-101-000",
    },
    "n": {
        "smallcaps": "000-501-161-106-000",
        "uppercase": "501-151-161-106-000",
        "lowercase": "000-525-101-101-000",
    },
    "o": {
        "smallcaps": "000-425-101-637-000",
        "uppercase": "425-101-101-637-000",
        "lowercase": "000-425-101-637-000",
    },
    "p": {
        "smallcaps": "000-126-137-100-000",
        "uppercase": "125-101-120-100-000",
        "lowercase": "000-525-101-137-100",
    },
    "q": {
        "smallcaps": "000-425-101-645-000",
        "uppercase": "425-101-101-645-000",
        "lowercase": "000-424-101-631-001",
    },
    "r": {
        "smallcaps": "000-125-137-101-000",
        "uppercase": "125-107-121-101-000",
        "lowercase": "000-575-100-100-000",
    },
    "s": {
        "smallcaps": "000-425-263-637-000",
        "uppercase": "425-630-001-637-000",
        "lowercase": "000-425-263-637-000",
    },
    "t": {
        "smallcaps": "000-212-010-010-000",
        "uppercase": "212-010-010-010-000",
        "lowercase": "040-212-010-063-000",
    },
    "u": {
        "smallcaps": "000-101-101-637-000",
        "uppercase": "101-101-101-637-000",
        "lowercase": "000-101-101-636-000",
    },
    "v": {
        "smallcaps": "000-101-607-010-000",
        "uppercase": "101-101-607-010-000",
        "lowercase": "000-101-607-010-000",
    },
    "w": {
        "smallcaps": "000-101-131-706-000",
        "uppercase": "101-101-131-706-000",
        "lowercase": "000-101-131-627-000",
    },
    "x": {
        "smallcaps": "000-637-010-425-000",
        "uppercase": "607-010-010-405-000",
        "lowercase": "000-637-010-425-000",
    },
    "y": {
        "smallcaps": "000-637-010-010-000",
        "uppercase": "101-637-010-010-000",
        "lowercase": "000-101-101-021-637",
    },
    "z": {
        "smallcaps": "000-267-010-453-000",
        "uppercase": "221-047-470-133-000",
        "lowercase": "000-267-010-453-000",
    },
    "*": {
        "smallcaps": "313-425-000-000-000",
        "uppercase": "313-425-000-000-000",
        "lowercase": "313-425-000-000-000",
    },
    "|": {
        "smallcaps": "1-1-1-1-1",
        "uppercase": "1-1-1-1-1",
        "lowercase": "1-1-1-1-1",
    },
    "!": {
        "smallcaps": "010-010-010-030-000",
        "uppercase": "010-010-010-030-000",
        "lowercase": "010-010-010-030-000",
    },
    "?": {
        "smallcaps": "425-007-040-030-000",
        "uppercase": "425-007-040-030-000",
        "lowercase": "425-007-040-030-000",
    },
    "@": {
        "smallcaps": "000-425-554-630-000",
        "uppercase": "000-425-554-630-000",
        "lowercase": "000-425-554-630-000",
    },
    "#": {
        "smallcaps": "000-355-407-662-000",
        "uppercase": "000-355-407-662-000",
        "lowercase": "000-355-407-662-000",
    },
    "&": {
        "smallcaps": "450-670-767-675-000",
        "uppercase": "450-670-767-675-000",
        "lowercase": "450-670-767-675-000",
    },
    "$": {
        "smallcaps": "010-462-615-357-010",
        "uppercase": "010-462-615-357-010",
        "lowercase": "010-462-615-357-010",
    },
    "€": {
        "smallcaps": "033-453-453-650-022",
        "uppercase": "033-453-453-650-022",
        "lowercase": "033-453-453-650-022",
    },
    "£": {
        "smallcaps": "030-425-430-400-763",
        "uppercase": "030-425-430-400-763",
        "lowercase": "030-425-430-400-763",
    },
    "¥": {
        "smallcaps": "607-313-313-010-010",
        "uppercase": "607-313-313-010-010",
        "lowercase": "607-313-313-010-010",
    },
    "~": {
        "smallcaps": "000-000-464-000-000",
        "uppercase": "000-000-464-000-000",
        "lowercase": "000-000-464-000-000",
    },
    "^": {
        "smallcaps": "030-425-000-000-000",
        "uppercase": "030-425-000-000-000",
        "lowercase": "030-425-000-000-000",
    },
    "_": {
        "smallcaps": "000-000-000-333-000",
        "uppercase": "000-000-000-333-000",
        "lowercase": "000-000-000-333-000",
    },
    "+": {
        "smallcaps": "000-010-212-020-000",
        "uppercase": "000-010-212-020-000",
        "lowercase": "000-010-212-020-000",
    },
    "-": {
        "smallcaps": "000-000-222-000-000",
        "uppercase": "000-000-222-000-000",
        "lowercase": "000-000-222-000-000",
    },
    "=": {
        "smallcaps": "000-333-333-000-000",
        "uppercase": "000-333-333-000-000",
        "lowercase": "000-333-333-000-000",
    },
    "/": {
        "smallcaps": "007-040-070-400-000",
        "uppercase": "007-040-070-400-000",
        "lowercase": "007-040-070-400-000",
    },
    "\\": {
        "smallcaps": "600-050-060-005-000",
        "uppercase": "600-050-060-005-000",
        "lowercase": "600-050-060-005-000",
    },
    "%": {
        "smallcaps": "207-040-070-403-000",
        "uppercase": "207-040-070-403-000",
        "lowercase": "207-040-070-403-000",
    },
    "<": {
        "smallcaps": "000-040-400-060-000",
        "uppercase": "000-040-400-060-000",
        "lowercase": "000-040-400-060-000",
    },
    ">": {
        "smallcaps": "000-050-005-070-000",
        "uppercase": "000-050-005-070-000",
        "lowercase": "000-050-005-070-000",
    },
    "(": {
        "smallcaps": "42-10-10-63-00",
        "uppercase": "42-10-10-63-00",
        "lowercase": "42-10-10-63-00",
    },
    ")": {
        "smallcaps": "25-01-01-37-00",
        "uppercase": "25-01-01-37-00",
        "lowercase": "25-01-01-37-00",
    },
    "[": {
        "smallcaps": "12-10-10-13-00",
        "uppercase": "12-10-10-13-00",
        "lowercase": "12-10-10-13-00",
    },
    "]": {
        "smallcaps": "21-01-01-31-00",
        "uppercase": "21-01-01-31-00",
        "lowercase": "21-01-01-31-00",
    },
    "{": {
        "smallcaps": "07-40-60-05-00",
        "uppercase": "07-40-60-05-00",
        "lowercase": "07-40-60-05-00",
    },
    "}": {
        "smallcaps": "60-05-07-40-00",
        "uppercase": "60-05-07-40-00",
        "lowercase": "60-05-07-40-00",
    },
    "`": {
        "smallcaps": "30-25-00-00-00",
        "uppercase": "30-25-00-00-00",
        "lowercase": "30-25-00-00-00",
    },
    ".": {
        "smallcaps": "0-0-0-3-0",
        "uppercase": "0-0-0-3-0",
        "lowercase": "0-0-0-3-0",
    },
    ",": {
        "smallcaps": "0-0-0-3-4",
        "uppercase": "0-0-0-3-4",
        "lowercase": "0-0-0-3-4",
    },
    ":": {
        "smallcaps": "0-3-0-3-0",
        "uppercase": "0-3-0-3-0",
        "lowercase": "0-3-0-3-0",
    },
    ";": {
        "smallcaps": "0-3-0-3-4",
        "uppercase": "0-3-0-3-4",
        "lowercase": "0-3-0-3-4",
    },
    "'": {
        "smallcaps": "5-0-0-0-0",
        "uppercase": "5-0-0-0-0",
        "lowercase": "5-0-0-0-0",
    },
    '"': {
        "smallcaps": "55-00-00-00-00",
        "uppercase": "55-00-00-00-00",
        "lowercase": "55-00-00-00-00",
    },
    "1": {
        "smallcaps": "000-310-010-313-000",
        "uppercase": "310-010-010-313-000",
        "lowercase": "310-010-010-313-000",
    },
    "2": {
        "smallcaps": "000-425-047-413-000",
        "uppercase": "425-007-040-453-000",
        "lowercase": "425-007-040-453-000",
    },
    "3": {
        "smallcaps": "000-425-025-637-000",
        "uppercase": "425-037-001-637-000",
        "lowercase": "425-037-001-637-000",
    },
    "4": {
        "smallcaps": "000-401-131-001-000",
        "uppercase": "401-101-221-001-000",
        "lowercase": "401-101-221-001-000",
    },
    "5": {
        "smallcaps": "000-122-225-637-000",
        "uppercase": "122-130-001-637-000",
        "lowercase": "122-130-001-637-000",
    },
    "6": {
        "smallcaps": "000-425-123-637-000",
        "uppercase": "425-130-101-637-000",
        "lowercase": "425-130-101-637-000",
    },
    "7": {
        "smallcaps": "000-221-047-010-000",
        "uppercase": "221-007-040-010-000",
        "lowercase": "221-007-040-010-000",
    },
    "8": {
        "smallcaps": "000-425-425-637-000",
        "uppercase": "425-637-101-637-000",
        "lowercase": "425-637-101-637-000",
    },
    "9": {
        "smallcaps": "000-425-231-637-000",
        "uppercase": "425-101-021-637-000",
        "lowercase": "425-101-021-637-000",
    },
    "0": {
        "smallcaps": "000-425-141-637-000",
        "uppercase": "425-141-171-637-000",
        "lowercase": "425-141-171-637-000",
    },
    "ç": {
        "smallcaps": "000-425-100-637-040",
        "uppercase": "000-425-100-637-040",
        "lowercase": "000-425-100-637-040",
    },
    "é": {
        "smallcaps": "003-042-000-425-122-637-000",
        "uppercase": "003-042-000-425-122-637-000",
        "lowercase": "003-042-000-425-122-637-000",
    },
    "è": {
        "smallcaps": "300-250-000-425-122-637-000",
        "uppercase": "300-250-000-425-122-637-000",
        "lowercase": "300-250-000-425-122-637-000",
    },
    "î": {
        "smallcaps": "030-425-000-210-010-313-000",
        "uppercase": "030-425-000-210-010-313-000",
        "lowercase": "030-425-000-210-010-313-000",
    },
    "•": {
        "smallcaps": "000-010-000-000-000",
        "uppercase": "000-010-000-000-000",
        "lowercase": "000-010-000-000-000",
    },
    "—": {
        "smallcaps": "000-000-222-000-000",
        "uppercase": "000-000-222-000-000",
        "lowercase": "000-000-222-000-000",
    },
}

block_palette = {
    "oak": {
        "block": Block("oak_planks"),
        "slab": Block("oak_slab"),
        "stairs": Block("oak_stairs"),
    },
    "spruce": {
        "block": Block("spruce_planks"),
        "slab": Block("spruce_slab"),
        "stairs": Block("spruce_stairs"),
    },
    "birch": {
        "block": Block("birch_planks"),
        "slab": Block("birch_slab"),
        "stairs": Block("birch_stairs"),
    },
    "jungle": {
        "block": Block("jungle_planks"),
        "slab": Block("jungle_slab"),
        "stairs": Block("jungle_stairs"),
    },
    "acacia": {
        "block": Block("acacia_planks"),
        "slab": Block("acacia_slab"),
        "stairs": Block("acacia_stairs"),
    },
    "dark_oak": {
        "block": Block("dark_oak_planks"),
        "slab": Block("dark_oak_slab"),
        "stairs": Block("dark_oak_stairs"),
    },
    "crimson": {
        "block": Block("crimson_planks"),
        "slab": Block("crimson_slab"),
        "stairs": Block("crimson_stairs"),
    },
    "warped": {
        "block": Block("warped_planks"),
        "slab": Block("warped_slab"),
        "stairs": Block("warped_stairs"),
    },
    "stone": {
        "block": Block("stone"),
        "slab": Block("stone_slab"),
        "stairs": Block("stone_stairs"),
    },
    "granite": {
        "block": Block("granite"),
        "slab": Block("granite_slab"),
        "stairs": Block("granite_stairs"),
    },
    "polished_granite": {
        "block": Block("polished_granite"),
        "slab": Block("polished_granite_slab"),
        "stairs": Block("polished_granite_stairs"),
    },
    "diorite": {
        "block": Block("diorite"),
        "slab": Block("diorite_slab"),
        "stairs": Block("diorite_stairs"),
    },
    "polished_diorite": {
        "block": Block("polished_diorite"),
        "slab": Block("polished_diorite_slab"),
        "stairs": Block("polished_diorite_stairs"),
    },
    "andesite": {
        "block": Block("andesite"),
        "slab": Block("andesite_slab"),
        "stairs": Block("andesite_stairs"),
    },
    "polished_andesite": {
        "block": Block("polished_andesite"),
        "slab": Block("polished_andesite_slab"),
        "stairs": Block("polished_andesite_stairs"),
    },
    "cobblestone": {
        "block": Block("cobblestone"),
        "slab": Block("cobblestone_slab"),
        "stairs": Block("cobblestone_stairs"),
    },
    "mossy_cobblestone": {
        "block": Block("mossy_cobblestone"),
        "slab": Block("mossy_cobblestone_slab"),
        "stairs": Block("mossy_cobblestone_stairs"),
    },
    "stone_brick": {
        "block": Block("stone_bricks"),
        "slab": Block("stone_brick_slab"),
        "stairs": Block("stone_brick_stairs"),
    },
    "mossy_stone_brick": {
        "block": Block("mossy_stone_bricks"),
        "slab": Block("mossy_stone_brick_slab"),
        "stairs": Block("mossy_stone_brick_stairs"),
    },
    "brick": {
        "block": Block("bricks"),
        "slab": Block("brick_slab"),
        "stairs": Block("brick_stairs"),
    },
    "end_stone_brick": {
        "block": Block("end_stone_bricks"),
        "slab": Block("end_stone_brick_slab"),
        "stairs": Block("end_stone_brick_stairs"),
    },
    "nether_brick": {
        "block": Block("nether_bricks"),
        "slab": Block("nether_brick_slab"),
        "stairs": Block("nether_brick_stairs"),
    },
    "red_nether_brick": {
        "block": Block("red_nether_bricks"),
        "slab": Block("red_nether_brick_slab"),
        "stairs": Block("red_nether_brick_stairs"),
    },
    "sandstone": {
        "block": Block("sandstone"),
        "slab": Block("sandstone_slab"),
        "stairs": Block("sandstone_stairs"),
    },
    "smooth_sandstone": {
        "block": Block("smooth_sandstone"),
        "slab": Block("smooth_sandstone_slab"),
        "stairs": Block("smooth_sandstone_stairs"),
    },
    "red_sandstone": {
        "block": Block("red_sandstone"),
        "slab": Block("red_sandstone_slab"),
        "stairs": Block("red_sandstone_stairs"),
    },
    "smooth_red_sandstone": {
        "block": Block("smooth_red_sandstone"),
        "slab": Block("smooth_red_sandstone_slab"),
        "stairs": Block("smooth_red_sandstone_stairs"),
    },
    "quartz": {
        "block": Block("quartz_block"),
        "slab": Block("quartz_slab"),
        "stairs": Block("quartz_stairs"),
    },
    "smooth_quartz": {
        "block": Block("smooth_quartz"),
        "slab": Block("smooth_quartz_slab"),
        "stairs": Block("smooth_quartz_stairs"),
    },
    "purpur": {
        "block": Block("purpur_block"),
        "slab": Block("purpur_slab"),
        "stairs": Block("purpur_stairs"),
    },
    "prismarine": {
        "block": Block("prismarine"),
        "slab": Block("prismarine_slab"),
        "stairs": Block("prismarine_stairs"),
    },
    "prismarine_brick": {
        "block": Block("prismarine_brick"),
        "slab": Block("prismarine_brick_slab"),
        "stairs": Block("prismarine_brick_stairs"),
    },
    "dark_prismarine": {
        "block": Block("dark_prismarine"),
        "slab": Block("dark_prismarine_slab"),
        "stairs": Block("dark_prismarine_stairs"),
    },
    "blackstone": {
        "block": Block("blackstone"),
        "slab": Block("blackstone_slab"),
        "stairs": Block("blackstone_stairs"),
    },
    "polished_blackstone": {
        "block": Block("polished_blackstone"),
        "slab": Block("polished_blackstone_slab"),
        "stairs": Block("polished_blackstone_stairs"),
    },
    "polished_blackstone_brick": {
        "block": Block("polished_blackstone_bricks"),
        "slab": Block("polished_blackstone_brick_slab"),
        "stairs": Block("polished_blackstone_brick_stairs"),
    },
}



from typing import Union

from shulker.components.Block import Block
from shulker.components.Zone import Zone
from shulker.components.BlockHandler import BlockHandler

from shulker.functions.base_functions import *


def meta_set_zone(
    zone: Zone, block: Block, handler: BlockHandler, filter: Union[Block, None] = None
) -> dict:

    instructions = {"list": []}

    if str(handler) == "replace" and filter:
        handler = str(handler) + " " + str(filter)

    instructions["list"].append(f"fill {zone} {block} {handler}")

    return instructions


def set_zone(
    zone: Union[Zone, list],
    block: Union[Block, str],
    handler: Union[BlockHandler, str] = "replace",
    filter: Union[Block, str] = "",
) -> str:
    """
    Returns a bool that is set to True
    if no message was sent back by the game or the
    message itself if there was an issue

    Available handlers:
        'replace' — The old block drops neither itself nor any contents. Plays no sound.
        'destroy' — The old block drops both itself and its contents (as if destroyed by a player). Plays the appropriate block breaking noise.
        'keep' — Only air blocks are changed (non-air blocks are unchanged).
        'hollow' - Replaces only the blocks on the outer edge of the fill region with the
        specified block. Inner blocks are changed to air, dropping their contents as entities
        but not themselves. If the fill region has no inner blocks (because it is smaller than
        three blocks in at least one dimension), acts like 'replace'.
        'outline' - Replaces only the blocks on the outer edge of the fill region with the
        specified block. Inner blocks are not affected. If the fill region has no inner blocks
        (because it is smaller than three blocks in at least one dimension), acts like 'replace'.

    Defaults to 'replace'
    """

    check_output_channel()

    zone = format_arg(zone, Zone)
    block = format_arg(block, Block)
    handler = format_arg(handler, BlockHandler)

    if filter != "":
        filter = format_arg(filter, Block)

    instructions = meta_set_zone(zone, block, handler, filter)

    for line in instructions["list"]:
        status = post(line)

    return default_check(status)



import unittest

from shulker.components.Entity import Entity, IncorrectEntityDescriptorType
from shulker.components.TargetSelector import TargetSelector


class TestBlock(unittest.TestCase):
    def test_entity_player_name(self):
        entity = Entity("Steve")

        diff = str(entity)
        test = "Steve"
        self.assertEqual(diff, test)

    def test_entity_UUID(self):
        entity = Entity("dd12be42-52a9-4a91-a8a1-11c01849e498")

        diff = str(entity)
        test = "dd12be42-52a9-4a91-a8a1-11c01849e498"
        self.assertEqual(diff, test)

    def test_entity_TargetSelector(self):
        ts = TargetSelector("p", {"tag": "cat"})
        entity = Entity(ts)

        diff = str(entity)
        test = "@p[tag=cat]"
        self.assertEqual(diff, test)

    def test_entity_invalid_descriptor(self):
        with self.assertRaises(IncorrectEntityDescriptorType):
            Entity(["apple", "dog"])

    def test_entity_incorrect_descriptor(self):
        with self.assertRaises(ValueError):
            Entity("")


if __name__ == "__main__":
    unittest.main()



import unittest

from shulker.components.BlockHandler import BlockHandler, BlockHandlerWrongType
from shulker.components.Block import Block


class TestBlockHandler(unittest.TestCase):
    def test_handler_destroy(self):
        bh = BlockHandler("destroy")

        diff = str(bh)
        test = "destroy"
        self.assertEqual(diff, test)

    def test_handler_replace(self):
        bh = BlockHandler("replace")

        diff = str(bh)
        test = "replace"
        self.assertEqual(diff, test)

    def test_handler_keep(self):
        bh = BlockHandler("keep")

        diff = str(bh)
        test = "keep"
        self.assertEqual(diff, test)

    def test_handler_hollow(self):
        bh = BlockHandler("hollow")

        diff = str(bh)
        test = "hollow"
        self.assertEqual(diff, test)

    def test_handler_outline(self):
        bh = BlockHandler("outline")

        diff = str(bh)
        test = "outline"
        self.assertEqual(diff, test)

    def test_handler_default(self):
        bh = BlockHandler()

        diff = str(bh)
        test = "replace"
        self.assertEqual(diff, test)

    def test_abritrary_handler(self):
        bh = BlockHandler("imaginary")

        with self.assertRaises(BlockHandlerWrongType):
            str(bh)

    def test_wrong_handler_type(self):
        bh = BlockHandler(1)

        with self.assertRaises(BlockHandlerWrongType):
            str(bh)


if __name__ == "__main__":
    unittest.main()



import urllib.request, json
import subprocess

class Runner():
    def __init__(self, verbose=False):
        self.jar_name = 'server.jar'
        self.version_manifest = 'https://launchermeta.mojang.com/mc/game/version_manifest.json'
        self.verbose = verbose
        self.stdout = subprocess.DEVNULL if not self.verbose else 1
        self.stderr = subprocess.DEVNULL if not self.verbose else 2

    def run(self):
        print("Browsering minecraft versions...")
        latest_url = self.browse_mc_versions()
        print("Getting latest version...")
        latest_version_info = self.get_jar_from_url(latest_url)
        print("Downloading latest version...")
        self.download_jar_and_generate(latest_version_info)
        print("Parsing jar...")
        self.parse_jar()
        print("Moving .json...")
        self.move_json()
        print("Cleaning up...")
        self.clean_up()

    def clean_up(self):
        instructions = [
            f'rm -rf {self.jar_name}',
            f'rm -rf logs',
            f'rm -rf generated',
            f'rm -rf libraries',
            f'rm -rf test',
            f'rm -rf versions'
        ]
        for instruction in instructions:
            subprocess.run(instruction.split(" "), stdout=self.stdout, stderr=self.stdout)
    
    def move_json(self):
        instructions = [
            f'mv ./generated/reports/blocks.json ../functions/mc_data/block_list.json',
            f'mv ./generated/reports/commands.json ../functions/mc_data/commands.json',
            f'mv ./generated/reports/registries.json ../functions/mc_data/registries.json',
        ]
        for instruction in instructions:
            subprocess.run(instruction.split(" "), stdout=self.stdout, stderr=self.stdout)
        
    def parse_jar(self):
        cmd = 'java -DbundlerMainClass=net.minecraft.data.Main -jar server.jar --all'.split(" ")
        subprocess.run(cmd, stdout=self.stdout, stderr=self.stdout)

    def download_jar_and_generate(self, url):
        download = f'wget {url} -O {self.jar_name}'.split(" ")
        subprocess.run(download, stdout=self.stdout, stderr=self.stdout)

    def get_jar_from_url(self, latest_url):
        with urllib.request.urlopen(latest_url) as url:
            latest_version_info = json.loads(url.read().decode())
            return latest_version_info['downloads']['server']['url']

    def browse_mc_versions(self):
        with urllib.request.urlopen(self.version_manifest) as url:
            server_jar_version_list = json.loads(url.read().decode())
            latest_version = server_jar_version_list["latest"]["release"]
            
            for version in server_jar_version_list["versions"]:
                if version['id'] == latest_version:
                    latest_url = version['url']
                    return latest_url

if __name__ == '__main__':
    tasks = Runner(verbose=True)
    tasks.run()




import re


class NBT:
    """
    Every arguments or attributes given to NBT must be written
    using the correct case provided by Minecraft.

    E.G.:
        -> AbsorptionAmount
        -> CustomName
        -> Fire

    If the case is not respected, the tags won't be applied.
    """

    def __init__(self, compound: dict = None):

        if isinstance(compound, dict):
            for key in compound:
                setattr(self, key, compound[key])

    def flatten(self, arg):
        if isinstance(arg, dict):
            return str(NBT(arg))

        elif isinstance(arg, str):  # see test_NBT_with_args in test_NBT.py
            if '"' in arg:
                return f"'{arg}'"
            else:
                return f'"{arg}"'

        elif isinstance(arg, float):
            return f"{arg}d"

        else:
            return str(arg)

    def __str__(self):

        buff = ""

        for key in dir(self):
            if key.startswith("__"):
                continue
            elif key == "flatten":
                continue

            value = getattr(self, key)
            value = self.flatten(value)

            buff += f"{key}:{value},"

        if buff != "":
            buff = buff[:-1]
            return f"{{{buff}}}"
        else:
            return ""
